module LazyME where

-- Some Haskell 98 features are not yet implemented in Hugs 98.
-- Features related to Rational are still to be implemented.
import Ratio

-- type Digit      = Integer
type Exponent   = Integer
type Mantissa   = [Integer]
type Exact      = (Exponent, Mantissa)

type Vector     = (Integer, Integer)
type Interval   = (Rational, Rational)
type Matrix     = (Vector,Vector)
type Tensor     = (Matrix,Matrix)
data Lft        = LftV Vector | LftM Matrix | LftT Tensor Integer
                deriving (Eq, Show)

data Expression = ExprV Vector
                | ExprI Interval
                | ExprM Matrix Expression
                | ExprT Tensor Integer Expression Expression
                deriving (Eq, Show)

{--

computeRBCut              :: Integer -> Expression -> [Tensor] -> Interval
computeRBCut prec arg ts  = foldRB x tList
  where
  x     = ((1,0),(0,1))
  tList = take (fromInteger cD) ts
  cD    = 3
--}

cValues         :: Integer -> (Integer -> Integer) -> (Integer -> Integer)
                -> (Integer,Integer)
cValues k fL fR = (toInteger cD, cX)
  where
  cD            = length (takeWhile (>(-k)) rightAccList)
  cX            = k + maximum( zipWith (+) leftList (0:rightAccList) )
  leftList      = take cD( map fL [0..] )
  rightAccList  = scanl1 (+) (map fR [0..])


foldRB                    :: Matrix -> [Tensor] -> Interval
foldRB x ts               = foldr fun (-1,1) ts
  where
  fun t intvl = tinfo2 (t `tleftm` x) (-1,1) intvl



mant2md :: Mantissa -> Matrix
mant2md ms = ((1,0),(k,2^(length ms)))
  where
  k = foldl1 (\x y -> 2*x+y) ms

-- exact2interval :: Integer -> Exact -> Interval
-- exact2interval k (e,ms) = 2^^e * (mantissaToRational (take (k+e) ms))
exact2interval k (e,ms) = minfo( mdotm (expM e) mpart )
  where
  mpart = mant2md (take (fromInteger (k+e)) ms)

testEM1 :: Exact
testEM1 = (3, cycle [1,1,1,1,0,0,0,1])
-- 7.56078431372549
testEM2 :: Exact
testEM2 = (0, cycle [1,1,1,1,0,0,0,1])
-- 0.945098039215686
testEM3 :: Exact
testEM3 = (0, cycle [0,0,0,1,1,1,1,0,0,0,1])
-- 0.117733268197362

testEM5 :: Exact
testEM5 = (0, repeat 0)
-- 0
testEM6 :: Exact
testEM6 = (0, repeat 1)
-- 1
testEM7 :: Exact
testEM7 = (0, 1 : (repeat 0) )
-- 1/2
testEM8 :: Exact
testEM8 = (0, repeat (-1) )
-- -1
testEM9 :: Exact
testEM9 = (0, (-1) : (repeat 0) )
-- -1/2

testEM10 :: Exact
testEM10 = (0, [0,0,0,0,0] ++ (repeat 1) )
-- 1/16


-- xExpEM :: Integer -> Exact -> Exact
xExpEM k (e,ms)
  | e <= 0    = _xExpA k (replicate (fromInteger (-e)) 0 ++ ms)
  | otherwise = tinfo2 tMul int1 int2
  where
  (ze,zms) = expr2EM (_xExpQ z 1)
  z        = foldl1 (\x y -> 2*x+y) (take (fromInteger e) ms)
  alpha    = drop (fromInteger e) ms
  int1     = exact2interval (k+ze+2) (ze,zms)
  int2     = _xExpA (k+ze+2) alpha

_xExpA :: Integer -> Mantissa -> Interval
_xExpA k ms = foldRB x ts
  where
  x        = mant2md( take (fromInteger cX) ms )
  ts       = map _ExpTensor [0..cD]
  (cD,cX)  = cValues k _expCUppL _expCUppR

myxExpA :: Integer -> Mantissa -> (Integer, Integer) -- Interval
myxExpA k ms = (cD,cX)
  where
  x        = mant2md( take (fromInteger cX) ms )
  ts       = map _ExpTensor [0..cD]
  (cD,cX)  = cValues k _expCUppL _expCUppR

_xExpQ :: Integer -> Integer -> Expression
_xExpQ p q     = ExprM _mFirst (eiterate _mat nC)
  where
  nC      = ceiling( (2*(abs p)-1) % q )
  _mFirst = mscale( foldl1 mdotm [ _mat i | i<-[0..(nC-1)] ] )
  _mat 0  = ((p,0),(p+q,q))
  _mat n  = ((p,0),(p,q*(n+1)))

_expCUppL   :: Integer -> Integer
_expCUppL 0 = 3
_expCUppL n = toInteger( ceiling( 3-(logBase 2 (fromInteger n + 1)) ) )

_expCUppR   :: Integer -> Integer
_expCUppR 0 = 2
_expCUppR n = toInteger( ceiling( 2-(logBase 2 (fromInteger n + 1)) ) )


{--
--------------------------------------
-- Exponential
--------------------------------------
eExp :: Expression -> Expression
eExp x        = eExpEM (expr2EM x)

eExpEM :: Exact -> Expression
eExpEM (e,ms)
  | e <= 0    = eExpA (exact2e (e,ms))
  | otherwise = (eExpQ z 1) `eMul` (eExpA alpha)
  where
  z     = foldl1 (\x y -> 2*x+y) (take (fromInteger e) ms)
  alpha = mant2e (drop (fromInteger e) ms)

eExpQ :: Integer -> Integer -> Expression
eExpQ p q     = ExprM _mFirst (eiterate _mat nC)
  where
  nC      = ceiling( (2*(abs p)-1) % q )
  _mFirst = mscale( foldl1 mdotm [ _mat i | i<-[0..(nC-1)] ] )
  _mat 0  = ((p,0),(p+q,q))
  _mat n  = ((p,0),(p,q*(n+1)))

eExpA :: Expression -> Expression
eExpA         = eiteratex _ExpTensor 0

_ExpTensor :: Integer -> Tensor
_ExpTensor 0  = (((1,0),(1,0)),((0,0),(1,1)))
_ExpTensor n  = (((1,0),(1,0)),((0,0),(0,n+1)))
--}


{--
xx :: Matrix -> [Tensor] -> Tensor
xx x ts = foldl fun idM (reverse ts)
  where
  fun t m = t `trightm` m
--}


{--
rbCut :: Integer -> Integer -> Mantissa -> [Tensor] -> Expression
rbCut cx cd ms ts = _rbCut2 x tens
  where
  x    = mant2expr (take (fromInteger cx) ms)
  tens = take (fromInteger cd) ts


_rbCut2 :: Expression -> [Tensor] -> Expression
_rbCut2 x []     = ExprI (-1,1)
_rbCut2 x (t:ts) = ExprT t 0 x (_rbCut2 x ts)


e2i                                   :: Expression -> Interval
e2i (ExprV v)                         = vinfo v
e2i (ExprI i)                         = i
e2i (ExprM m (ExprI i))               = (p,q)
  where
  list = [matrixX m (fst i), matrixX m (snd i)]
  p = minimum list
  q = maximum list
e2i (ExprM m (ExprV v))               = vinfo (m `mdotv` v)
e2i (ExprM m x)                       = e2i (ExprM m (ExprI (e2i x)))
e2i (ExprT t i (ExprV vx) (ExprV vy)) = vinfo (t `tleftv` vx `mdotv` vy)
e2i (ExprT t i (ExprV vx) y)          = e2i (ExprM (t `tleftv` vx) y)
e2i (ExprT t i x (ExprV vy))          = e2i (ExprM (t `trightv` vy) x)
e2i (ExprT t i (ExprI ix) (ExprI iy)) = (p,q)
  where
  list = [tensorXY t (fst ix) (fst iy), tensorXY t (fst ix) (snd iy),
          tensorXY t (snd ix) (fst iy), tensorXY t (snd ix) (snd iy)]
  p = minimum list
  q = maximum list
e2i (ExprT t i x y)                   = e2i (ExprT t i (ExprI ix) (ExprI iy))
  where
  ix = e2i x
  iy = e2i y


interval2mat i = mscale ((r-p,s-q),(r+p,s+q))
  where
  (p,q) = (numerator (fst i), denominator (fst i))
  (r,s) = (numerator (snd i), denominator (snd i))
--}

{--
xxxxx_emApprox :: Integer -> Exact -> Interval
xxxxx_emApprox n (e,ms) = minfo _mat
  where
  _mat = foldl mdotm (expM e) (map digM (take (fromInteger n) ms))
--}


emCut :: Integer -> Exact -> Exact
emCut n (e,ms) = (e,take (fromInteger n) ms)

superset :: Interval -> Interval -> Bool
superset (a,b) (c,d) = (a<=c) && (b>=d)

containsEM :: Interval -> Exact -> Bool
containsEM (p,q) (e,ms) = superset (p,q) (minfo _mat)
  where
  _mat = foldl mdotm (expM e) (map digM ms)

-- Returns index of succesful test
rif :: Exact -> [Exact -> Bool] -> Int
rif = _rif2 0

_rif2 :: Integer -> Exact -> [Exact -> Bool] -> Int
_rif2 n x fs
  | len1 == len2  = _rif2 (n+1) x fs
  | otherwise     = len2
  where
  xCut  = emCut n x
  len1  = length fs
  len2  = length tests
  tests = takeWhile not (map (\f -> f xCut) fs)

eem = sem
expr2EM = sem
mant2e = mant2expr
emitDigits = dem

--------------------------------------
-- Normalisation
--------------------------------------
couldNormalise         :: Exact -> (Bool, Exact)
couldNormalise (e, []) = (True, (e,[]))
couldNormalise (e,[0]) = (True, (e-1,[]))
couldNormalise (e,[m]) = (True, (e,[m]))
couldNormalise (e,(m1:m2:ms))
  | e <= zeroFlag       = (False, (e,m1:m2:ms))
  | m1 == 0             = couldNormalise (e-1,m2:ms)
  | (m1,m2) == (1,-1)   = couldNormalise (e-1, 1:ms)
  | (m1,m2) == (-1,1)   = couldNormalise (e-1,-1:ms)
  | otherwise           = (True, (e,m1:m2:ms))


-------------------------------------------------------------------------------
-- Basic Functions for Exact Type
-------------------------------------------------------------------------------


normalise             :: Exact -> Exact
normalise (e, [])   = (e,[])
normalise (e,[0])   = (e-1,[])
normalise (e,[m])   = (e,[m])
normalise (e,(m1:m2:ms))
--  | e <= zeroFLAG     = (e,m1:m2:ms,1)
  |  m1 == 0          = normalise (e-1,m2:ms)
  | (m1,m2) == (1,-1) = normalise (e-1, 1:ms)
  | (m1,m2) == (-1,1) = normalise (e-1,-1:ms)
  | otherwise         = (e,m1:m2:ms)

zeroFlag              :: Integer
zeroFlag              = -1000

{--
maybeZero             :: Exact -> Bool
maybeZero (_,_,f)     = f==1

normalise             :: Exact -> Exact
normalise (e, [],f)   = (e,[],f)
normalise (e,[0],f)   = (e-1,[],f)
normalise (e,[m],f)   = (e,[m],f)
normalise (e,(m1:m2:ms),f)
  | e <= zeroFLAG     = (e,m1:m2:ms,1)
  |  m1 == 0          = normalise (e-1,m2:ms,f)
  | (m1,m2) == (1,-1) = normalise (e-1, 1:ms,f)
  | (m1,m2) == (-1,1) = normalise (e-1,-1:ms,f)
  | otherwise         = (e,m1:m2:ms,f)
--}


--------------------------------------------------------------------------------
-- Basic Functions
--------------------------------------------------------------------------------
one                   :: a -> Integer -> a
one f 1               = f

idM                   :: Matrix
idM                   = ((1,0),(0,1))

trans                 :: ((t1,t2),(t3,t4)) -> ((t1,t3),(t2,t4))
trans ((a,b),(c,d))   = ((a,c),(b,d))

inverse               :: Matrix -> Matrix
inverse ((a,b),(c,d)) = mscale ((d,-b),(-c,a))

--------------------------------------------------------------------------------
-- Binary Scaling Functions
--------------------------------------------------------------------------------
vscale                      :: Vector -> Vector
vscale (a,b)
    | ar && br              = vscale (div a 2, div b 2)
    | otherwise             = (a,b)
    where
    ar = (mod a 2 == 0)
    br = (mod b 2 == 0)

mscale                      :: Matrix -> Matrix
mscale ((a,b),(c,d))
    | ar && br && cr && dr  = mscale ((div a 2, div b 2), (div c 2, div d 2))
    | otherwise             = ((a,b),(c,d))
    where
    ar = (mod a 2 == 0)
    br = (mod b 2 == 0)
    cr = (mod c 2 == 0)
    dr = (mod d 2 == 0)

tscale                      :: Tensor -> Tensor
tscale (((a,b),(c,d)),((e,f),(g,h)))
    | ar && br && cr && dr &&
      er && fr && gr && hr  = tscale (((div a 2, div b 2), (div c 2, div d 2)),
                                      ((div e 2, div f 2), (div g 2, div h 2)))
    | otherwise             = (((a,b),(c,d)),((e,f),(g,h)))
    where
    ar = (mod a 2 == 0)
    br = (mod b 2 == 0)
    cr = (mod c 2 == 0)
    dr = (mod d 2 == 0)
    er = (mod e 2 == 0)
    fr = (mod f 2 == 0)
    gr = (mod g 2 == 0)
    hr = (mod h 2 == 0)

--------------------------------------------------------------------------------
-- Exponent and Digit Matrices (and correpsonding Lft inverses)
--------------------------------------------------------------------------------
expM          :: Integer -> Matrix
expM ee
  | ee >= 0   = ((2^ee,0),(0,1 ))
  | otherwise = ((1,0),(0,2^(-ee)))

iexpM         :: Integer -> Lft
iexpM ee
  | ee >= 0   = LftM ((1,0),(0,2^ee))
  | otherwise = LftM ((2^(-ee),0),(0,1))

digM          :: Integer -> Matrix
digM k        = ((1,0),(k,2))

multidigM     :: Integer -> Integer -> Matrix
multidigM k n = ((1,0),(k,2^n))

dneg          :: Matrix
dzer          :: Matrix
dpos          :: Matrix
dneg          = ((1,0),(-1,2))
dzer          = ((1,0),( 0,2))
dpos          = ((1,0),( 1,2))

idneg         :: Lft
idzer         :: Lft
idpos         :: Lft
idneg         = LftM (inverse dneg)
idzer         = LftM (inverse dzer)
idpos         = LftM (inverse dpos)

-------------------------------------------------------------------------------
-- matrixX, tensorXY
-------------------------------------------------------------------------------
matrixX                 :: Matrix -> Rational -> Rational
matrixX ((a,b),(c,d)) x = (ar*x+cr) / (br*x+dr)
  where
  ar = fromInteger a
  br = fromInteger b
  cr = fromInteger c
  dr = fromInteger d

tensorXY                :: Tensor -> Rational -> Rational -> Rational
tensorXY (((a,b),(c,d)),((e,f),(g,h))) x y
                        = (ar*x*y+cr*x+er*y+gr) / (br*x*y+dr*x+fr*y+hr)
  where
  ar = fromInteger a
  br = fromInteger b
  cr = fromInteger c
  dr = fromInteger d
  er = fromInteger e
  fr = fromInteger f
  gr = fromInteger g
  hr = fromInteger h

--------------------------------------------------------------------------------
-- Tensors for Basic Arithmetic Operations
--------------------------------------------------------------------------------
tAdd                :: Tensor
tSub                :: Tensor
tMul                :: Tensor
tDiv                :: Tensor
tAdd                = (((0,0),(1,0)),((1,0),(0,1)))
tSub                = (((0,0),(1,0)),((-1,0),(0,1)))
tMul                = (((1,0),(0,0)),((0,0),(0,1)))
tDiv                = (((0,0),(1,0)),((0,1),(0,0)))

--------------------------------------------------------------------------------
-- Linear Fractional Transformation Products
--------------------------------------------------------------------------------
mdotv                       :: Matrix -> Vector -> Vector
mdotv ((a,b),(c,d)) (e,f)   = (a * e + c * f,b * e + d * f)

mdotm                       :: Matrix -> Matrix -> Matrix
mdotm m (v,w)               = (mdotv m v,mdotv m w)

mdott                       :: Matrix -> Tensor -> Tensor
mdott m (n,o)               = (mdotm m n,mdotm m o)

tleftv                      :: Tensor -> Vector -> Matrix
tleftv t v                  = trightv (trans t) v

tleftm                      :: Tensor -> Matrix -> Tensor
tleftm t m                  = trans (trightm (trans t) m)

trightv                     :: Tensor -> Vector -> Matrix
trightv (m,n) v             = (mdotv m v,mdotv n v)

trightm                     :: Tensor -> Matrix -> Tensor
trightm (m,n) o             = (mdotm m o,mdotm n o)

dot                         :: Integer -> Lft -> Lft -> Lft
dot 1 (LftM m) (LftV v)     = LftV (vscale (mdotv m v))
dot 1 (LftM m) (LftM n)     = LftM (mscale (mdotm m n))
dot 1 (LftM m) (LftT t i)   = LftT (tscale (mdott m t)) i
dot 1 (LftT t i) (LftV v)   = LftM (mscale (tleftv t v))
dot 1 (LftT t i) (LftM m)
    | m == idM              = LftT t i
    | otherwise             = LftT (tscale (tleftm t m)) (i+1)
dot 2 (LftT t i) (LftV v)   = LftM (mscale (trightv t v))
dot 2 (LftT t i) (LftM m)
    | m == idM              = LftT t i
    | otherwise             = LftT (tscale (trightm t m)) (i+1)

--------------------------------------------------------------------------------
-- Tests for Boundness and Refinement
--------------------------------------------------------------------------------
bounded                       :: Lft -> Bool
bounded (LftV v)              = vbounded v
bounded (LftM m)              = mbounded m
bounded (LftT t i)            = tbounded t

vbounded                      :: Vector -> Bool
vbounded (a,b)                = ( b /= 0 )

mbounded                      :: Matrix -> Bool
mbounded ((_,b),(_,d))        = ( abs(d) > abs(b) )

tbounded                      :: Tensor -> Bool
tbounded (((_,b),(_,d)),((_,f),(_,h)))
                              = (cond1 && cond2 && cond3 && cond4)
  where
  cond1 = abs(h) > abs(b+d+f)
  cond2 = abs(h) > abs(b+d-f)
  cond3 = abs(h) > abs(b-d+f)
  cond4 = abs(h) > abs(b-d-f)

refine                        :: Lft -> Bool
refine (LftV v)               = vrefine v
refine (LftM m)               = mrefine m
refine (LftT t i)             = trefine t

vrefine                       :: Vector -> Bool
vrefine (a,b)                 = (vbounded (a,b)) && (abs(a) <= abs(b))

mrefine                       :: Matrix -> Bool
mrefine ((a,b),(c,d))         = (mbounded ((a,b),(c,d))) && cond1 && cond2
  where
  cond1 = abs(c-a) <= abs(d-b)
  cond2 = abs(c+a) <= abs(d+b)

trefine                       :: Tensor -> Bool
trefine (((a,b),(c,d)),((e,f),(g,h)))
                              = cond1 && cond2 && cond3 && cond4 && cond5
  where
  cond1 = tbounded (((a,b),(c,d)),((e,f),(g,h)))
  cond2 = abs( a+c+e+g) <= abs( b+d+f+h)
  cond3 = abs(-a-c+e+g) <= abs(-b-d+f+h)
  cond4 = abs(-a+c-e+g) <= abs(-b+d-f+h)
  cond5 = abs( a-c-e+g) <= abs( b-d-f+h)

-------------------------------------------------------------------------------
-- Information
-------------------------------------------------------------------------------
info                          :: Lft -> (Rational,Rational)
info (LftV v)                 = vinfo v
info (LftM m)                 = minfo m
info (LftT t i)               = tinfo t

vinfo                         :: Vector -> (Rational,Rational)
vinfo (a,b)
  | not(vbounded (a,b))       = error "Error 101 (infoV) - vector not bounded"
  | otherwise                 = (a%b,a%b)

minfo                         :: Matrix -> (Rational,Rational)
minfo ((a,b),(c,d))
  | not(mbounded ((a,b),(c,d)))
                              = error "Error 103 (infoM) - matrix not bounded"
  | s ==  1                   = (p,q)
  | s == -1                   = (q,p)
  | otherwise                 = (p,p)
  where
  s = signum(a*d-b*c)
  p = (c-a)%(d-b)
  q = (c+a)%(d+b)

minfo2                        :: Matrix -> Interval -> Interval
minfo2 m x                    = (left,right)
  where
  left    = minimum values
  right   = maximum values
  values  = [matrixX m x1, matrixX m x2]
  (x1,x2) = x


tinfo                         :: Tensor -> (Rational,Rational)
tinfo t
  | not(tbounded t)           = error ("Error 104 " ++ (show t))
--  | not(tbounded t)           = error "Error 104 (infoT) - tensor not bounded"
  | otherwise                 = (left,right)
  where
  left   = minimum values
  right  = maximum values
  values = [ tensorXY t (-1) (-1), tensorXY t (-1) 1,
             tensorXY t   1  (-1), tensorXY t   1  1 ]

tinfo2                        :: Tensor -> Interval -> Interval -> Interval
tinfo2 t x y                  = (left,right)
  where
  left    = minimum values
  right   = maximum values
  values  = [ tensorXY t x1 y1, tensorXY t x1 y2,
             tensorXY t x2 y1, tensorXY t x2 y2 ]
  (x1,x2) = x
  (y1,y2) = y


--------------------------------------------------------------------------------
-- Radius
--------------------------------------------------------------------------------
radius                :: Lft -> Exponent
radius l              = if (2^^e >= r) then e else (e+1)
  where
  i = info l
  r = max (abs (fst i)) (abs (snd i))
  e = ceiling (logBase 2 (fromRational r))

--------------------------------------------------------------------------------
-- LFT numerator and denominator
--------------------------------------------------------------------------------
lftNum                                        :: Lft -> Lft
lftNum (LftM ((a,b),(c,d)))                   = LftM ((a,0),(c,1))
lftNum (LftT (((a,b),(c,d)),((e,f),(g,h))) i) = LftT (((a,0),(c,0)),((e,0),(g,1))) i

lftDen                                        :: Lft -> Lft
lftDen (LftM ((a,b),(c,d)))                   = LftM ((b,0),(d,1))
lftDen (LftT (((a,b),(c,d)),((e,f),(g,h))) i) = LftT (((b,0),(d,0)),((f,0),(h,1))) i

--------------------------------------------------------------------------------
-- Square Bracket Application
--------------------------------------------------------------------------------
app              :: Lft -> (Integer -> Expression) -> Expression
app (LftV v) g   = ExprV v
app (LftM m) g   = cons (dot 1 (LftM m) (hd (g 1))) (tl (g 1))
app (LftT t i) g = cons (dot 1 (dot 2 (LftT t i) (hd (g 2))) (hd (g 1))) h
  where
  c = branch (hd (g 1))
  h i
    | i <= c     = tl (g 1) i
    | otherwise  = tl (g 2) (i-c)







--------------------------------------------------------------------------------
-- Tensor Absorption Strategy
--------------------------------------------------------------------------------
{--
vlessv                      :: Vector -> Vector -> Bool
vlessv v w                  = determinant (v,w) < 0

mlessv                      :: Matrix -> Vector -> Bool
mlessv (v,w) x              = vlessv v x && vlessv w x

mlessm                      :: Matrix -> Matrix -> Bool
mlessm m (v,w)              = mlessv m v && mlessv m w

mdisjointm                  :: Matrix -> Matrix -> Bool
mdisjointm m n              = mlessm m n || mlessm n m
--}

{--
strategyf                   :: Tensor -> Integer -> Integer
strategyf t i               = mod i 2 + 1

strategyo                   :: Tensor -> Integer -> Integer
strategyo t
    | trefine t             = strategyr t
    | otherwise             = strategyf t


-- strategyr                   :: Tensor -> Integer -> Integer
-- strategyr t i
--     | overlap (minfo m1) (minfo m2) = 1
--     | otherwise                     = 2
--     where
--     m1 = fst (trans t)
--     m2 = snd (trans t)

strategyr                   :: Tensor -> Integer -> Integer
strategyr t i
    | mdisjointm t1 t2      = 2
    | otherwise             = 1
    where
    t1 = fst (trans t)
    t2 = snd (trans t)

decision                    :: Integer -> Lft -> Bool
decision 1 (LftM m)         = True
decision 1 (LftT t i)       = strategyo t i == 1
decision 2 (LftT t i)       = strategyo t i == 2

overlap :: Interval -> Interval -> Bool
overlap (a,b) (c,d) = (b>=c) && (d>=a)
--}

-------------
-- alternate
-------------
decision                    :: Integer -> Lft -> Bool
decision 1 (LftM m)         = True
decision 1 (LftT t i)       = mod i 2 == 0
decision 2 (LftT t i)       = mod i 2 == 1












--------------------------------------------------------------------------------
-- Basic Expression Tree Functions
--------------------------------------------------------------------------------
branch                  :: Lft -> Integer
branch (LftV v)         = 0
branch (LftM m)         = 1
branch (LftT t i)       = 2

vis                     :: Lft -> Bool
vis (LftV v)            = True
vis e                   = False

mis                     :: Lft -> Bool
mis (LftM m)            = True
mis e                   = False

tis                     :: Lft -> Bool
tis (LftT t i)          = True
tis e                   = False

hd                      :: Expression -> Lft
hd (ExprV v)            = LftV v
hd (ExprM m e)          = LftM m
hd (ExprT t i e f)      = LftT t i

tl                      :: Expression -> Integer -> Expression
tl (ExprM m e) 1        = e
tl (ExprT t i e f) 1    = e
tl (ExprT t i e f) 2    = f

zerCol                  :: Vector -> Bool
zerCol (a,b)            = a==0 && b==0

cons                            :: Lft -> (Integer -> Expression) -> Expression
cons (LftV v) f                 = ExprV v
cons (LftM (v1,v2)) f
  | zerCol v1                   = ExprV v2
  | otherwise                   = ExprM (v1,v2) (f 1)
cons (LftT ((v1,v2),(v3,v4)) i) f
  | zerCol v1 && zerCol v2 && zerCol v3
                                = ExprV v4
  | zerCol v1 && zerCol v2      = ExprM (v3,v4) (f 2)
  | zerCol v1 && zerCol v3      = ExprM (v2,v4) (f 1)
  | otherwise                   = ExprT ((v1,v2),(v3,v4)) i (f 1) (f 2)
--  | otherwise                   = app (LftT ((v1,v2),(v3,v4)) i) f

--------------------------------------------------------------------------------
-- Normalization Functions
--------------------------------------------------------------------------------
sem                 :: Expression -> Exact
sem e
  | bounded (hd e2) = (resE, dem resMs)
  | otherwise       = sem2 e2
  where
  resE  = radius (hd e2) -- TRY: max 0 (radius (hd e2))
  resMs = app (iexpM resE) (one e2)
  g j = tl e2 j
  e2  = app l f
  l   = hd e
  f d = ExprM (expM r) (cons (dot 1 (iexpM r) bh) h)
    where
    bh  = hd (tl e d)
    r   = radius bh        -- TRY: max 0 (radius bh)
    h k = tl (tl e d) k

sem2 e = emDiv (sem en) (sem ed)
{--
-- = error (show e) -- sem en -- error "Hi there" -- (2,[])
  | maybeZero y    = error ("1 - " ++ (show y))
  | otherwise      = error ("2 - " ++ (show y)) -- "Hi there"
--}
  where
  y   = normalise (sem ed)
  en  = cons (lftNum l) f
  ed  = cons (lftDen l) f
  l   = hd e
  f d = tl e d

dem                                 :: Expression -> Mantissa
dem e
    | refine (dot 1 idneg l)        = -1 : (dem (app idneg (one e)))
    | refine (dot 1 idpos l)        =  1 : (dem (app idpos (one e)))
    | refine (dot 1 idzer l)        =  0 : (dem (app idzer (one e)))
    | otherwise                     = dem (app l f)
    where
    l = hd e
    f d = ab l (tl e d) (decision d l)

ab                                  :: Lft -> Expression -> Bool -> Expression
ab k e b
    | not b                         = ExprM idM e
    | tis k && tis (hd e)           = mant2expr (dem e)
    | otherwise                     = e

--------------------------------------------------------------------------------
-- Type Cast Functions
--------------------------------------------------------------------------------
exact2e             :: Exact -> Expression
exact2e (e,ms)      = ExprM (expM e) (mant2expr ms)

mant2expr              :: Mantissa -> Expression
mant2expr []           = ExprI (-1,1)
-- mant2expr []           = ExprM idM (mant2expr [])
mant2expr (m:ms)       = ExprM (digM m) (mant2expr ms)

--------------------------------------------------------------------------------
-- Decimal Output Function
--------------------------------------------------------------------------------
printExact          :: Exact -> Int -> String
-- printExact (e,ms,f) n = if (f==1) then "maybe ZERO" else show (fromRational res)
printExact (e,ms) n = show (fromRational res)
  where
  res = 2^^e * (mantissaToRational (take n ms))

printMantissa         :: Mantissa -> Int -> String
printMantissa ms n    = show (fromRational res)
  where
  res = mantissaToRational (take n ms)

mantissaToRational    :: Mantissa -> Rational
mantissaToRational ms = (toInteger num) % den
  where
  num = foldl (\x y -> 2*x+y) 0 ms
  den = 2^(length ms)

--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
-- getResultDigits :: Integer -> Exact -> Matrix
-- getResultDigits n 


--------------------------------------------------------------------------------
-- Simple Functions on Exacts
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Constructing Expression Tree
--------------------------------------------------------------------------------
eiterate            :: (Integer -> Matrix) -> Integer -> Expression
eiterate i n        = ExprM (i n) (eiterate i (n+1))

eiteratex           :: (Integer -> Tensor) -> Integer -> Expression -> Expression
eiteratex i n x     = ExprT (i n) 0 x (eiteratex i (n+1) x)

--------------------------------------------------------------------------------
-- Pi, e
--------------------------------------------------------------------------------
{--
epi                 :: Exact -- Expression
epi                 = sem ( ExprT tdiv 0 (exact2e (eSqrtQ 10005 1)) eomega )

eomega              :: Expression
eomega              = eiterate iteromega 0
iteromega           :: Integer -> Matrix
iteromega 0         = ((6795705,213440),(6795704,213440))
iteromega n         = ((e-d-c,e+d+c),(e+d-c,e-d+c))
    where
    b = (2*n-1)*(6*n-5)*(6*n-1)
    c = b*(545140134*n + 13591409)
    d = b*(n+1)
    e = 10939058860032000*n^4


a0 = a13 % a42
a1 = - (a54+a13) % a42
ax = 1 % a64^3
iM1 = (( a1*ax, 0), (a0+a1*ax,1))
--}

ePi           :: Expression
ePi           = ExprT tDiv 0 (exact2e (emSqrtQ 10005 1)) eOmega

eOmega        :: Expression
eOmega        = eiterate eOmegaItere 0

eOmegaItere 0 = ((-558731543,0),
                 (3568253353002447403380457,112071970708091043840000))
eOmegaItere n = mscale ((a,0),(a,b))
  where
  a = (6*n+1)*(6*n+2)*(6*n+3)*(6*n+4)*(6*n+5)*(6*n+6)*(a54*(n+1)+a13)
  b = -(n+1)^3*(3*n+1)*(3*n+2)*(3*n+3)*(a54*n+a13)*a64^3
  a13 = 13591409
  a42 = 426880
  a54 = 545140134
  a64 = 640320

ee                  :: Expression
ee                  = eiterate eeItere 0

eeItere             :: Integer -> Matrix
eeItere 0           = ((1,1),(6,2))
eeItere n           = ((0,1),(4,8*n+4))




eLn2 = ln2

ln2                 :: Expression
ln2                 = eiterate ln2Itere 0

ln2Itere            :: Integer -> Matrix
ln2Itere 0          = ((1,2),(4,6))
ln2Itere n          = ((-n,2*n+1),(-4*n,7*n+3))

cuppr 0 = 2
cuppr n = ceiling( logBase 2 (8 / ((n+1)*(2*n+3))) )
cUPPR n = foldl1 (+) (map cuppr [0..n])
-- cAPPR :: Integer -> Integer
cAPPR n = (cuppr 0) + 5*n - 1 - (floor aux)
  where
  nF = fromInteger n
  aux = 2*nF*(logBase 2 nF) + (logBase 2 (2*pi*nF))



--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------






eexpszer            :: Expression -> Expression
eexpszer            = eiteratex iterexpszer 0

iterexpszer         :: Integer -> Tensor
iterexpszer n       = (((2*n+2,2*n+1),(2*n+1,2*n)),
                       ((2*n,2*n+1),(2*n+1,2*n+2)))



--------------------------------------------------------------------------------
-- Some Examples
--------------------------------------------------------------------------------
ex0 = ExprT tAdd 0 (ExprM ((1,0),(0,2)) (ExprV (3,2))) (ExprV (3,2))
ex1 = ExprM ((1,0),(0,4)) (ExprV (3,2))

ex10 = ExprT tAdd 0 (ExprM ((1,0),(0,2)) (ExprV (1,2))) (ExprV (7,5))
ex11 = ExprT (((0,0),(1,0)),((4,0),(0,8))) 2 (ExprM ((1,0),(0,1)) (ExprV (1,2))) (ExprV (7,10))
ex12 = ExprT tSub 0 (ExprM ((1,0),(0,2)) (ExprV (1,2))) (ExprV (7,5))
ex13 = ExprT tDiv 0 (ExprM ((1,0),(0,2)) (ExprV (1,2))) (ExprV (7,5))
ex14 = ExprT tMul 0 (ExprM ((1,0),(0,2)) (ExprV (1,2))) (ExprV (7,5))


t0 = (((0,0),(1,0)),((1,1),(0,1)))
t1 = (((1,0),(1,1)),((3,1),(0,3)))
t2 = (((0,0),(1,0)),((4,1),(0,4)))
t3 = (((0,0),(1,0)),((2,1),(0,2)))

s0 = (((0,0),(1,0)),((1,1),(0,1)))
s1 = trightm (mdott (inverse dpos) s0) dpos
s2 = trightm (mdott (inverse dzer) s1) dzer
s3 = trightm (mdott (inverse dpos) s1) dpos




--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

--------------------------------------
-- Negation
--------------------------------------
eNeg :: Expression -> Expression
eNeg x       = app (LftM ((-1,0),(0,1))) (one x)

emNeg :: Exact -> Exact
emNeg (e,ms) = (e, map negate ms)

--------------------------------------
-- Absolute Value
--------------------------------------
emAbs :: Exact -> Exact
emAbs (e,ms) = (e,_absM ms)
  where
  _absM []      = []
  _absM ( 0:xs) = 0 : (_absM xs)
  _absM ( 1:xs) = 1 : xs
  _absM (-1:xs) = 1 : (map negate xs)

--------------------------------------
-- Addition and Subtraction
--------------------------------------
eAdd :: Expression -> Expression -> Expression
eAdd x y = cons (LftT tAdd 0) f
  where
  f 1 = x
  f 2 = y

emAdd :: Exact -> Exact -> Exact
emAdd (e1,ms1) (e2,ms2) = (e+1,ms)
  where
  e    = max e1 e2
  ms   = emitDigits (cons (LftT _ten 0) g)
  g 1  = mant2e (replicate (fromInteger (e-e1)) 0 ++ ms1)
  g 2  = mant2e (replicate (fromInteger (e-e2)) 0 ++ ms2)
  _ten = (((0,0),(1,0)),((1,0),(0,2)))

eSub :: Expression -> Expression -> Expression
eSub x y = cons (LftT tSub 0) f
  where
  f 1 = x
  f 2 = y

emSub :: Exact -> Exact -> Exact
emSub (e1,ms1) (e2,ms2) = (e+1,ms)
  where
  e    = max e1 e2
  ms   = emitDigits (cons (LftT _ten 0) g)
  g 1  = mant2e (replicate (fromInteger (e-e1)) 0 ++ ms1)
  g 2  = mant2e (replicate (fromInteger (e-e2)) 0 ++ ms2)
  _ten = (((0,0),(1,0)),((-1,0),(0,2)))

--------------------------------------
-- Multiplication and Division
--------------------------------------
eMul :: Expression -> Expression -> Expression
eMul x y                = cons (LftT tMul 0) f
  where
  f 1 = x
  f 2 = y

emMul :: Exact -> Exact -> Exact
emMul (e1,ms1) (e2,ms2) = (e1+e2,ms)
  where
  ms  = emitDigits (cons (LftT tMul 0) f)
  f 1 = mant2e ms1
  f 2 = mant2e ms2 

emRec :: Exact -> Exact
emRec y
  | isOK        = (-e+2, emitDigits b2)
  | otherwise   = error "Possible division by zero"
  where
  (isOK,(e,m1:m2:ms)) = couldNormalise y
  b2 = cons (LftM r2) (one (mant2e ms))
  r2 = r `mdotm` (digM m1) `mdotm` (digM m2)
  r  = ((0,4),(1,0))

emDiv :: Exact -> Exact -> Exact
emDiv x y = x `emMul` (emRec y)

eDiv :: Expression -> Expression -> Expression
eDiv x y = exact2e( (expr2EM x) `emDiv` (expr2EM y) )

--------------------------------------
-- Exponential
--------------------------------------
eExp :: Expression -> Expression
eExp x        = eExpEM (expr2EM x)

eExpEM :: Exact -> Expression
eExpEM (e,ms)
  | e <= 0    = eExpA (exact2e (e,ms))
  | otherwise = (eExpQ z 1) `eMul` (eExpA alpha)
  where
  z     = foldl1 (\x y -> 2*x+y) (take (fromInteger e) ms)
  alpha = mant2e (drop (fromInteger e) ms)

eExpQ :: Integer -> Integer -> Expression
eExpQ p q     = ExprM _mFirst (eiterate _mat nC)
  where
  nC      = ceiling( (2*(abs p)-1) % q )
  _mFirst = mscale( foldl1 mdotm [ _mat i | i<-[0..(nC-1)] ] )
  _mat 0  = ((p,0),(p+q,q))
  _mat n  = ((p,0),(p,q*(n+1)))

eExpA :: Expression -> Expression
eExpA         = eiteratex _ExpTensor 0

_ExpTensor :: Integer -> Tensor
_ExpTensor 0  = (((1,0),(1,0)),((0,0),(1,1)))
_ExpTensor n  = (((1,0),(1,0)),((0,0),(0,n+1)))


--------------------------------------
-- Natural Logarithm
--------------------------------------
-- eLn :: Expression -> Expression

{--
emMul                         :: Exact -> Exact -> Exact
emMul (e1,ms1) (e2,ms2) = normalise (e1+e2,ms)
  where
  ms = dem (cons (LftT tmul 0) g)
  g 1 = mant2expr ms1
  g 2 = mant2expr ms2 
--}

emLn x
  | _isOK        = res3 -- res1 `emAdd` res2 -- ((ExprV (e-1,1)) `eMul` eLn2) `eAdd` (eLnA alpha)
  | otherwise    = error "Invalid argument?"
  where
  res3  = mant2e (tail ms) -- (e,ms) -- alpha
  res1  = expr2EM (eLnA alpha)
  res2  = expr2EM (ExprV (e-1,1) `eMul` eLn2)
  alpha = mant2e (tail ms)
  (_isOK,(e,ms)) = couldNormalise (emAbs x)

alph :: Expression
alph = mant2e ( tail (cycle [1,0,1,-1]) )

eLnA :: Expression -> Expression
eLnA          = eiteratex _LnTensor 0

_LnTensor :: Integer -> Tensor
_LnTensor 0   = (((1,0),(4,2)),((0,1),(0,4)))
_LnTensor n   = (((2*n,0),(8*n,2*n+2)),((0,2*n+1),(0,8*n+4)))



--------------------------------------
-- Cosine
--------------------------------------
-- eCos :: Expression -> Expression
eCos x = _eCosF alpha
  where
--  (e,ms) = expr2EM (x `eDiv` twoPi)
--  twoPi  = app (LftM ((2,0),(0,1))) (one ePi)
  (e,ms) = expr2EM (x `eMul` twoPiInverse)
  twoPiInverse  = app (LftM ((0,2),(1,0))) (one ePi)
  alpha
    | e>=0      = drop (fromInteger e) ms
    | otherwise = (replicate (fromInteger (-e) ) 0) ++ ms


--
-- in 0:0:ms
-- the first 0 is because x^2 \in [0,4] and tensors have already
--   absorbed the exponent matrix E_2
-- the second 0 is because in eCosA we must multiply the
--   argument with pi/2. We mulitply it with pi only.
--
_eCosF :: Mantissa -> Expression
_eCosF (-1:  0 :ms) = eNeg (eCosA (mant2expr (0:0:ms))) -- error "-COS ms"
_eCosF (-1:  1 :ms) =      (eSinA (mant2expr (0:0:ms))) -- error "+SIN ms"
_eCosF ( 0:(-1):ms) =      (eSinA (mant2expr (0:0:ms))) -- error "+SIN ms"
_eCosF ( 0:  0 :ms) =      (eCosA (mant2expr (0:0:ms))) -- error "+COS ms"
_eCosF ( 0:  1 :ms) = eNeg (eSinA (mant2expr (0:0:ms))) -- error "-SIN ms"
_eCosF ( 1:(-1):ms) = eNeg (eSinA (mant2expr (0:0:ms))) -- error "-SIN ms"
_eCosF ( 1:  0 :ms) = eNeg (eCosA (mant2expr (0:0:ms))) -- error "-COS ms"
_eCosF ( 1:  1 :ms) =      (eSinA (mant2expr (0:0:ms))) -- error "+SIN ms"


eCosA :: Expression -> Expression
eCosA x        = eiteratex _CosTensor 0 (y `eMul` y)
  where
  y = x `eMul` ePi

_CosTensor :: Integer -> Tensor
_CosTensor 0   = (((-2,0),(-2,0)),((0,0),(1,1)))
_CosTensor n   = (((-2,0),(-2,0)),((0,0),(0,(n+1)*(2*n+1))))






-- eCo2s :: Expression -> Expression
eCos2 x = _eCosF2 alpha
  where
  (e,ms) = expr2EM x -- should be (e,ms) = x / 2Pi
  alpha
    | e>=0      = drop (fromInteger e) ms
    | otherwise = (replicate (fromInteger (-e) ) 0) ++ ms

_eCosF2 :: Mantissa -> Expression
_eCosF2 (-1:(-1):ms) = eNeg (eSinA2 (mant2expr ms)) -- error "-SIN ms"
_eCosF2 (-1:  0 :ms) = eNeg (eCosA2 (mant2expr ms)) -- error "-COS ms"
_eCosF2 (-1:  1 :ms) =      (eSinA2 (mant2expr ms)) -- error "+SIN ms"
_eCosF2 ( 0:(-1):ms) =      (eSinA2 (mant2expr ms)) -- error "+SIN ms"
_eCosF2 ( 0:  0 :ms) =      (eCosA2 (mant2expr ms)) -- error "+COS ms"
_eCosF2 ( 0:  1 :ms) = eNeg (eSinA2 (mant2expr ms)) -- error "-SIN ms"
_eCosF2 ( 1:(-1):ms) = eNeg (eSinA2 (mant2expr ms)) -- error "-SIN ms"
_eCosF2 ( 1:  0 :ms) = eNeg (eCosA2 (mant2expr ms)) -- error "-COS ms"
_eCosF2 ( 1:  1 :ms) =      (eSinA2 (mant2expr ms)) -- error "+SIN ms"


eCosA2 :: Expression -> Expression
eCosA2 x        = eiteratex _CosTensor2 0 (x `eMul` x)

_CosTensor2 :: Integer -> Tensor
_CosTensor2 0   = (((-1,0),(-1,0)),((0,0),(2,2)))
_CosTensor2 n   = (((-1,0),(-1,0)),((0,0),(0,2*(n+1)*(2*n+1))))




eSinA2 x = error "eCosA2"
-- eSinA2 _ = ExprV (1,1)

eSinA x = error "eCosA"
-- eSinA _ = ExprV (1,1)













--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

--------------------------------------
-- Square Root
--------------------------------------
_SqrtTensor = (((0,0),(4,0)),((1,1),(3,7)))
id2pos = LftM ((4,0),(-3,1))

emSqrt :: Exact -> Exact
emSqrt y
  | not (isOK)  = error "Could not normalise"
  | even e      = (div e 2, _sqrtMs ms)
  | odd e       = (div (e+1) 2, _sqrtMs (0:ms))
  where
  (isOK,(e,ms)) = couldNormalise y

_sqrtTests = [ containsEM ( 1%2,   1),
               containsEM ( 3%8, 5%8),
               containsEM ( 1%4, 1%2),
               containsEM (3%16,5%16),
               containsEM ( 1%8, 1%4) ]

_sqrtMs ms = snd (eliminateExp res)
  where
  i = rif (0,ms) _sqrtTests
  y = case i of
    0 -> _sqrtAlpha ( mant2e ms )
    1 -> _sqrtAlpha ( (ExprV (3,2)) `eMul` (mant2e ms) )
    2 -> _sqrtAlpha ( (ExprV (2,1)) `eMul` (mant2e ms) )
    3 -> _sqrtAlpha ( (ExprV (3,1)) `eMul` (mant2e ms) )
    4 -> _sqrtAlpha ( (ExprV (4,1)) `eMul` (mant2e ms) )
  res = case i of
    0 -> expr2EM y
    1 -> expr2EM ( (eSqrtQ 2 3) `eMul` y )
    2 -> expr2EM ( (eSqrtQ 2 1) `eMul` y )
    3 -> expr2EM ( (eSqrtQ 3 1) `eMul` y )
    4 -> expr2EM ( (eSqrtQ 4 1) `eMul` y )

-- We know the argument is in [1/2,1]
_sqrtAlpha a = ExprM dpos (ExprM dpos res) -- (fbLoop _SqrtTensor ms2)
  where
  res = fbl _SqrtTensor (cons id2pos (one a))
--  ms2 = emitDigits( cons id2pos (one a) )

fbl :: Tensor -> Expression -> Expression
fbl t x
  | trefine tP   = ExprM dpos ( fbl tP x )
  | trefine tN   = ExprM dneg ( fbl tN x )
  | trefine tZ   = ExprM dzer ( fbl tZ x )
  | otherwise    = fbl _ten (app l f)
  where
  tP   = tscale( (inverse dpos) `mdott` t `trightm` dpos )
  tN   = tscale( (inverse dneg) `mdott` t `trightm` dneg )
  tZ   = tscale( (inverse dzer) `mdott` t `trightm` dzer )
  l    = hd x
  f d  = ab l (tl x d) (decision d l)
  LftT _ten _ = dot 1 (LftT t 0) l


eliminateExp (e,ms) = pushExp e (e,ms)

pushExp 0 (e,ms)        = (e,ms)
pushExp i (e,[])        = (e,[])
pushExp i (e,0:ms)      = pushExp (i-1) (e-1,ms)
pushExp i (e,1:(-1):ms) = pushExp (i-1) (e-1,1:ms)
pushExp i (e,(-1):1:ms) = pushExp (i-1) (e-1,-1:ms)
pushExp 1 (e,m:0:ms)    = pushExpAux (e,m:0:ms)
pushExp _ _               = error "Cannot push any more"

pushExpAux (e,m:0:ms)
--  | f2 == 1          = error "I'm not sure if the push is possible"
--  | ms2 == []        = error "Push not possible"
--  | head ms2 == m    = error "Sorry, cannot push"
  | otherwise        = (e-1,msNew)
  where
  (e2,ms2) = normalise (0,0:ms)
  msNew = (replicate (fromInteger (-e2+1)) m) ++ (tail ms)
{--
pushExpAux (e,m:0:ms,f)
  | f2 == 1          = error "I'm not sure if the push is possible"
  | ms2 == []        = error "Push not possible"
  | head ms2 == m    = error "Sorry, cannot push"
  | otherwise        = (e-1,msNew,f)
  where
  (e2,ms2,f2) = normalise (0,0:ms,0)
  msNew = (replicate (fromInteger (-e2+1)) m) ++ (tail ms)
--}

{--
eSqrt :: Exact -> Exact
eSqrt (e,ms)
  | even e    = eSqrt2 (div e 2,     preProcess ms)
  | odd  e    = eSqrt2 (div (e+1) 2, preProcess (0:ms))
--}

sqrtT = (((0,0),(4,0)),((1,1),(3,7)))

-- postProcess [0] = [-1]
-- postProcess ms  = 1 : (map (\ 0 -> (-1)) ms)

eSqrt2 (e,0:0:ms)
--  | e <= 100             = (e, 1 : map (\0 -> (-1)) (0:0:ms), 0)
  | e <= zeroFlag        = (e, 1 : map (\0 -> (-1)) (0:0:ms))
  | otherwise            = eSqrt2 (e-1,ms)
eSqrt2 (e,1:1:ms)        = (e, 1:1: (fbLoop sqrtT ms))
eSqrt2 (e,1:0:1:ms)      = (e, 1:1: (fbLoop sqrtT (-1:ms)))
eSqrt2 (e,1:0:0:ms)      = (e, msNew)
  where
  (_,msNew)   = eliminateExp (emMul (emSqrtQ 2 3) (0,msTemp))
  (_,ms2)     = eliminateExp (emMul (0,1:0:0:ms) (sem (ExprV (3,2))))
  msTemp      = 1:1: (fbLoop sqrtT msTemp2)
  msTemp2     = case ms2 of
                1:1:ms3 -> ms3
                1:ms3   -> dem( app idpos (g ms3) )
                ms3     -> dem( app (LftM ((4,0),(-3,1))) (g ms3) )
  g ls k = mant2expr ls
eSqrt2 (e,1:0:(-1):ms)   = eSqrt2 (e,0:1:1:ms)
eSqrt2 (e,1:(-1):ms)     = eSqrt2 (e,0:1:ms)
eSqrt2 (e,0:1:1:ms)      = ( e, msNew)
  where
  (_,msNew) = eliminateExp (emMul (emSqrtQ 1 2) (0,msTemp))
  (_,ms2)   = eliminateExp (1,0:1:1:ms)
  msTemp      = 1:1: (fbLoop sqrtT msTemp2)
  msTemp2     = case ms2 of
                1:1:ms3 -> ms3
                1:ms3   -> dem( app idpos (g ms3) )
                ms3     -> dem( app (LftM ((4,0),(-3,1))) (g ms3) )
  g ls k = mant2expr ls
eSqrt2 (e,0:1:0:1:ms)    = eSqrt2 (e,0:1:1:(-1):ms)
eSqrt2 (e,0:1:0:0:ms)    = ( e, msNew)
  where
  (_,msNew) = eliminateExp (emMul (emSqrtQ 1 3) (0,msTemp))
  (_,ms2)   = eliminateExp (emMul (0,0:1:0:0:ms) (sem (ExprV (3,1))))
  msTemp      = 1:1: (fbLoop sqrtT msTemp2)
  msTemp2     = case ms2 of
                1:1:ms3 -> ms3
                1:ms3   -> dem( app idpos (g ms3) )
                ms3     -> dem( app (LftM ((4,0),(-3,1))) (g ms3) )
  g ls k = mant2expr ls
eSqrt2 (e,0:1:0:(-1):ms) = ( e-1, 1:1: (fbLoop sqrtT ms))
eSqrt2 _                   = error "Something gone wrong!"

preProcess              :: Mantissa -> Mantissa
preProcess []           = []
preProcess [m]          = [m]
preProcess ( 0:  ms)    = 0 : (preProcess ms)
preProcess ( 1:(-1):ms) = 0 : (preProcess (1:ms))
preProcess ( 1:  0 :ms) = 1 : 0 : ms
preProcess ( 1:  1 :ms) = 1 : 1 : ms
preProcess (-1:(-1):ms) = 1 : 1 : (map negate ms)
preProcess (-1:  0 :ms) = 1 : 0 : (map negate ms)
preProcess (-1:  1 :ms) = 0 : (preProcess (1:(map negate ms)))

--------------------------------------
-- Square Root
--------------------------------------
fbLoop :: Tensor -> Mantissa -> Mantissa
fbLoop _ []      = []
fbLoop t (m:ms)
  | trefine tP   =  1 : ( fbLoop tP (m:ms) )
  | trefine tN   = -1 : ( fbLoop tN (m:ms) )
  | trefine tZ   =  0 : ( fbLoop tZ (m:ms) )
  | otherwise    = fbLoop _ten ms
  where
  tP   = tscale( (inverse dpos) `mdott` t `trightm` dpos )
  tN   = tscale( (inverse dneg) `mdott` t `trightm` dneg )
  tZ   = tscale( (inverse dzer) `mdott` t `trightm` dzer )
  _ten = tscale( t `tleftm` (digM m) )

eSqrtQ :: Integer -> Integer -> Expression
eSqrtQ p q = exact2e (emSqrtQ p q)

emSqrtQ :: Integer -> Integer -> Exact
emSqrtQ p q
  | q >= p    = (0, _sqrtQ ((2*q,q),(p,2*q)))
  | otherwise = (e, _sqrtQ ((2*s,s),(p,2*s)))
  where
  r = radius (LftV (p,q))
  e = div r 2 + mod r 2
  s = q * 2^(2*e)

_sqrtQ :: Matrix -> Mantissa
_sqrtQ m
  | c >= 0         =  1 : _sqrtQ (f dpos)
  | otherwise      = -1 : _sqrtQ (f dneg)
  where
  f k = (inverse k) `mdotm` m `mdotm` k
  c   = (fst . snd) m

  
------------------------------------------------------
-- Example for Abbas
-- 17 March 2005
------------------------------------------------------
expLeft :: Integer -> Integer
expLeft 0 = 3
expLeft i = ceiling( logBase 2 (8 / fromInteger (i+1)) )

expLefts :: [Integer]
expLefts = 3 : (map f [1..])
  where
  f i = ceiling( logBase 2 (8 / (i+1)) )

expRight :: Integer -> Integer
expRight 0 = 2
expRight i = ceiling( logBase 2 (4 / fromInteger (i+1)) )

expRights :: [Integer]
expRights = 2 : (map f [1..])
  where
  f i = ceiling( logBase 2 (4 / fromInteger (i+1)) )

expDepth :: Integer -> Integer
expDepth k = toInteger ( length ( takeWhile (>0) (scanl (+) k expRights) ) - 1 )

expArgDig :: Integer -> Integer
expArgDig k = sum (take (fromInteger k) expRights) + (expLeft k)

expArgDigits :: [Integer]
expArgDigits = map expArgDig [0..]

expPiCF :: [Integer]
expPiCF = [3,7,15,1,292,1,1,1,2,1,3,1,14,2,1,1,2,2,2,2]

expPiMat :: Integer -> Matrix
expPiMat 0 = ((1,0),(3,1))
expPiMat k = ((0,1),(1, expPiCF!!(fromInteger k) ))

expPiApprox :: Integer -> Matrix
expPiApprox 0 = expPiMat 0
expPiApprox k = (expPiApprox (k-1)) `mdotm` (expPiMat k)


info2Double :: Interval -> (Double, Double)
info2Double (p,q) = (fromRational p, fromRational q)


