module Picard_11 where
import Picard_11_Data_Types
import Picard_11_Basics
import Picard_11_Update
import Picard_11_Graphical_Part
import Picard_11_Tests
import GraphicsUtils
import FastSort


--------------------------------------------------------------------------------
-- drawUpdates
--
-- drawUpdatesExample
--------------------------------------------------------------------------------
drawUpdatesExample         :: Int -> Int -> IO()
drawUpdatesExample ex iter = drawUpdates (pExVF ex) (pExFun ex) iter
--
--
drawUpdates                :: QCubes -> QSteps -> Int -> IO()
drawUpdates vfs fs i       = runGraphics ( do
  win1 <- openWindow "draw updates (in: vfs, fs)" (xWindow, yWindow)
  drawInWindow win1 grid
  drawInWindow win1 drawKey
  -- Initial/Input Functions
  drawInWindow win1 (drawQCubes  19 vfs)
  drawInWindow win1 (drawQSteps  11  fs)
  getKey win1
  -- Draw function updates
  aux_drawUpdates win1 [61,62..] (fList,gList)
  closeWindow win1
  )
  where
  (fList,gList) = unzip upList
  upList        = computeUpdates vfs fs i
--
--
aux_drawUpdates            :: Window -> [Int] -> ([QSteps],[QSteps]) -> IO()
aux_drawUpdates w (c:cs) ((f:fs),(g:gs)) = do
  modGraphic w (drawF `overGraphic`)
  directDraw w drawF
  getKey w
  directDraw w drawG
  getKey w
  case (length fs > 0) of
    True  -> aux_drawUpdates w cs (fs,gs)
    False -> putStr "That's all folks\n"
  where
  drawF = drawQSteps c f
  drawG = drawQSteps (c+100) g
--
--
--------------------------------------------------------------------------------
-- computeUpdates
--
--------------------------------------------------------------------------------
computeUpdates          :: QCubes -> QSteps -> Int -> [(QSteps,QSteps)]
computeUpdates vfs fs i = take i resList
  where
  resList = zip fList gList
  gList   = [ recs2Steps (updateDerivative vfs f) | f<-fList ]
  fList   = fs : [ updateFunction f g | f<-fList, g<-gList ]
--
--
--------------------------------------------------------------------------------
-- ud_draw
--
-- ud_drawExample
--------------------------------------------------------------------------------
ud_drawExample        :: Int -> IO()
ud_drawExample i      = ud_draw ( (ud_exampleVF i), (ud_exampleFun i) )
--
--
ud_draw               :: (QCubes,QSteps) -> IO()
ud_draw (vfs,fs)      = runGraphics ( do
  win1 <- openWindow "draw derivative update (in: vfs, fs)" (xWindow,yWindow)
  drawInWindow win1 grid
  drawInWindow win1 drawKey
  -- Initial Interval Functions
  drawInWindow win1 (drawQSteps  12 fs)
  getKey win1
  -- Draw vector field
  drawInWindow win1 (drawQCubes  19 vfs)
  getKey win1
  -- Draw derivative update
  -- drawInWindow win1 (drawQRecs  112 grs1)
  drawInWindow win1 (drawQSteps 151 gs1)
  getKey win1
  closeWindow win1
  )
  where
  grs1 = updateDerivative vfs fs
  gs1   = recs2Steps grs1
  fs1   = updateFunction fs gs1
--
--
--------------------------------------------------------------------------------
-- uf_drawS
--
-- uf_drawSExample
--------------------------------------------------------------------------------
uf_drawSExample       :: Int -> Int -> Int -> IO()
uf_drawSExample i k m = uf_drawS (uf_exampleQ i k m)
--
--
uf_drawS              :: (QSteps, QSteps) -> IO()
uf_drawS (fs,gs)      = runGraphics ( do
  win1 <- openWindow "draw function update (in: fs, gs)" (xWindow, yWindow)
  drawInWindow win1 grid
  drawInWindow win1 drawKey
  -- Initial Interval Functions
  drawInWindow win1 (drawQSteps  12 fs)
  drawInWindow win1 (drawQSteps 112 gs)
  getKey win1
  drawInWindow win1 (drawQSteps  30 fupdated)
  drawInWindow win1 (drawQSteps  40 fstrong)
  getKey win1
  closeWindow win1
  )
  where
  fupdated = updateFunction fs gs
  fstrong  = updateSCFunction fs gs
--
--
--------------------------------------------------------------------------------
-- uf_drawR
--
-- uf_drawRExample
--------------------------------------------------------------------------------
uf_drawRExample       :: Int -> Int -> Int -> IO()
uf_drawRExample i k m = uf_drawR (uf_exampleR i k m)
--
--
uf_drawR              :: (QRecs,QRecs) -> IO()
uf_drawR (frs,grs)    = runGraphics ( do
  win1 <- openWindow "draw Updates (in: fs, gs)" (xWindow, yWindow)
  drawInWindow win1 grid
  drawInWindow win1 drawKey
  -- Initial Functions as Recs
  drawInWindow win1 (drawQRecs   11 frs)
  drawInWindow win1 (drawQRecs  111 grs)
  -- Initial Interval Functions
  drawInWindow win1 (drawQSteps  12 fs)
  drawInWindow win1 (drawQSteps 112 gs)
  getKey win1
  -- left to right
  drawInWindow win1 (drawQSteps  33 f_L2R)
  -- drawInWindow win1 (drawQSteps  43 f_L2RSC)
  getKey win1
  -- right to left
  drawInWindow win1 (drawQSteps  34 f_R2L)
  -- drawInWindow win1 (drawQSteps  44 f_R2LSC)
  getKey win1
  -- function update, strong function update
  drawInWindow win1 (drawQSteps  30 fupdated)
  -- drawInWindow win1 (drawQSteps  40 fstrong)
  getKey win1
  closeWindow win1
  )
  where
  fs = recs2Steps frs
  gs = recs2Steps grs
  gsSC = flipLU2UL gs
  [fs2,gs2] = syncFunctions [fs,gs]
  f_L2R = uf_L2R fs2 gs2
  [fs3,gs3] = syncFunctions [f_L2R,gs]
  f_R2L = uf_R2L fs3 gs3
  [fs4,gs4] = syncFunctions [fs,gsSC]
  f_L2RSC = uf_L2R fs4 gs4
  [fs5,gs5] = syncFunctions [f_L2RSC,gs]
  f_R2LSC = uf_R2L fs5 gs5
  fupdated = updateFunction fs gs
  fstrong = updateSCFunction fs gs

