module Picard_11_Graphical_Part where
import Picard_11_Data_Types
import GraphicsUtils


type Q2Dpoint = (Rational,Rational)


drawQCubes                      :: Int -> QCubes -> Graphic
drawQCubes i cs                 = overGraphics (map (drawQCube i) cs)
--
--
drawQCube                       :: Int -> QCube -> Graphic
drawQCube i ((a,b),(c,d),(e,f)) = overGraphics [graph1,graph2]
  where
  graph1 = withRGB (rgb i) (recf (a,c) (b,d))
  graph2 = withRGB (rgb (i+colourFunDer)) (recg (a,e) (b,f))
--
--
-------------------------------------------------------------------------------
-- drawQRecs
-------------------------------------------------------------------------------
drawQRecs              :: Int -> QRecs -> Graphic
drawQRecs  i fs        = overGraphics (map (drawQRecs_1 i) fs)
--
drawQRecs_1            :: Int -> QRec -> Graphic
drawQRecs_1 i ((a,b),(c,d))
  | i < colourFunDer   = withRGB (rgb i) (recf (a,c) (b,d))
  | otherwise          = withRGB (rgb i) (recg (a,c) (b,d))
--
--
--------------------------------------------------------------------------------
-- drawQSteps, drawQStep
--------------------------------------------------------------------------------
drawQSteps           :: Int -> QSteps -> Graphic
drawQSteps i fs      = overGraphics (map (drawQStep i) fs)
--
drawQStep            :: Int -> QStep -> Graphic
drawQStep i (QIDef ((a,b),(c,d),(e,f)))
  | i < colourFunDer = withRGB (rgb i) (overGraphics
                       [linef (a,c) (b,d), linef (a,e) (b,f)])
  | otherwise        = withRGB (rgb i) (overGraphics
                       [lineg (a,c) (b,d), lineg (a,e) (b,f)])
drawQStep i (QINotDef (a,b))
  | i < colourFunDer = withRGB rgbNotDef (recFilledf (a,-ctyf) (b,ctyf))
  | otherwise        = withRGB rgbNotDef (recFilledg (a,-ctyg) (b,ctyg))
drawQStep i (QIFail (a,b))
  | i < colourFunDer = withRGB rgbFail (recFilledf (a,-ctyf) (b,ctyf))
  | otherwise        = withRGB rgbFail (recFilledg (a,-ctyg) (b,ctyg))
--
--
{--
-------------------------------------------------------------------------------
--
-------------------------------------------------------------------------------
drawQIFunction        :: Int -> QIFunction -> Graphic
drawQIFunction i fs   = overGraphics (map (drawQIFunction_1 i) fs)
--
drawQIFunction_1      :: Int -> QIFunction_1 -> Graphic
drawQIFunction_1 i (QIDef ((a,b),(c,d),(e,f)))
  | i < colourFunDer = withRGB (rgb i) (overGraphics
                       [linef (a,c) (b,d), linef (a,e) (b,f)])
  | otherwise        = withRGB (rgb i) (overGraphics
                       [lineg (a,c) (b,d), lineg (a,e) (b,f)])
drawQIFunction_1 i (QINotDef (a,b))
  | i < colourFunDer = withRGB rgbNotDef (recFilledf (a,-ctyf) (b,ctyf))
  | otherwise        = withRGB rgbNotDef (recFilledg (a,-ctyg) (b,ctyg))
drawQIFunction_1 i (QIFail (a,b))
  | i < colourFunDer = withRGB rgbFail (recFilledf (a,-ctyf) (b,ctyf))
  | otherwise        = withRGB rgbFail (recFilledg (a,-ctyg) (b,ctyg))
--
--
-------------------------------------------------------------------------------
--
-------------------------------------------------------------------------------
drawQCFunction        :: Int -> QCFunction -> Graphic
drawQCFunction i fs   = overGraphics (map (drawQCFunction_1 i) fs)
--
drawQCFunction_1      :: Int -> QCFunction_1 -> Graphic
drawQCFunction_1 i (QCDef ((a,b),(c,d)))
  | i < colourFunDer = withRGB (rgb i) (linef (a,c) (b,d))
  | otherwise        = withRGB (rgb i) (lineg (a,c) (b,d))
drawQCFunction_1 i (QCNotDef (a,b))
  | i < colourFunDer = withRGB rgbNotDef (recFilledf (a,-ctyf) (b,ctyf))
  | otherwise        = withRGB rgbNotDef (recFilledg (a,-ctyg) (b,ctyg))
drawQCFunction_1 i (QCFail (a,b))
  | i < colourFunDer = withRGB rgbFail (recFilledf (a,-ctyf) (b,ctyf))
  | otherwise        = withRGB rgbFail (recFilledg (a,-ctyg) (b,ctyg))
--
--
--}


-------------------------------------------------------------------------------
-- xWindow, yWindow   = dimensions (in pixels) of windows drawn
-- fyWindow, gyWindow = heights of 'function' and 'derivative' parts
-- partitionWindow    = height of the 'partition' part
-- xSpace, ySpace     = empty zone between graphs and/or borders
-- arrowSpace         = space for arrow tip
--
-- yfWindow + ygWindow + partitionWindow + 2*arrowSpace + 3*ySpace = yWindow
--
-- xMin, xMax         = range of the x coordinates (default interval [0,1])
-- yfMin, yfMax       = range of y-axis for functions
-- ygMin, ygMax       = range of y-axis for derivatives
-- yfUnit, ygUnit     = size (in pixels) of y-axis unit for function/derivative
-------------------------------------------------------------------------------
xWindow, yWindow :: Int
xWindow          =   700
yWindow          =   800
--
yfWindow, ygWindow, partitionWindow, xSpace, ySpace, arrowSpace :: Int
yfWindow         = 450
ygWindow         = 100
partitionWindow  =  50
xSpace           =  50
ySpace           =  50
arrowSpace       =  25
--
xMin, xMax       :: Int
xMin             =  -1
xMax             =   1   --  1
--
yfMin, yfMax     :: Int
yfMin            =   0   -- -1
yfMax            =   6   --  3
--
ygMin, ygMax     :: Int
ygMin            =   1   -- -1
ygMax            =   4   --  1
--
xUnit, yfUnit, ygUnit :: Int
xUnit            = div (xWindow - xSpace - arrowSpace) (xMax - xMin)
yfUnit           = div yfWindow (yfMax - yfMin)
ygUnit           = div ygWindow (ygMax - ygMin)
--
--
-------------------------------------------------------------------------------
-- cCf = change coordinates of an object drawn in the 'function part'
-- cCp = change coordinates of an object drawn in the 'partition part'
-- cCg = change coordinates of an object drawn in the 'derivative part'
-------------------------------------------------------------------------------
cCf, cCg, cCp :: Q2Dpoint -> Point
cCf (a,b) = (xSpace + x, arrowSpace + yfWindow - y)
  where
	x = round ( (a - fromInt xMin) * (fromInt xUnit) )
	y = round ( (b - fromInt yfMin) * (fromInt yfUnit) )
cCp (a,b) = (xSpace + x, arrowSpace + ySpace + yfWindow + partitionWindow - y)
  where
  x = round ( (a - fromInt xMin) * (fromInt xUnit) )
  y = round ( b * (fromInt ygUnit) )
cCg (a,b) = (xSpace + x, yWindow - ySpace - y)
  where
	x = round ( (a - fromInt xMin) * (fromInt xUnit) )
	y = round ( (b - fromInt ygMin) * (fromInt ygUnit) )
--
--
-------------------------------------------------------------------------------
-- rec = draws a rectangle; (a,b) - left lower corner
--                          (c,d) - right upper corner
-------------------------------------------------------------------------------
rec             :: Point -> Point -> Graphic
rec (a,b) (c,d) = polyline [(a,b), (c,b),(c,d), (a,d), (a,b)]
--
--
-------------------------------------------------------------------------------
-- recf, recFilledf, polygonf, linef, textf
--      = draw rec, recFilled, polygon, line and text in the 'function part'
-- recg, recFilledg, polygong, lineg, textg
--      = draw rec, recFilled, polygon, line and text in the 'derivative part'
-------------------------------------------------------------------------------
recf       :: Q2Dpoint -> Q2Dpoint -> Graphic
recf p1 p2 = rec (cCf p1) (cCf p2)
--
recg       :: Q2Dpoint -> Q2Dpoint -> Graphic
recg p1 p2 = rec (cCg p1) (cCg p2)
--
recFilledf             :: Q2Dpoint -> Q2Dpoint -> Graphic
recFilledf (a,b) (c,d) = polygonf [(a,b), (c,b), (c,d), (a,d), (a,b)]
--
recFilledg             :: Q2Dpoint -> Q2Dpoint -> Graphic
recFilledg (a,b) (c,d) = polygong [(a,b), (c,b), (c,d), (a,d), (a,b)]
--
polygonf :: [Q2Dpoint] -> Graphic
polygonf = polygon . (map cCf)
--
polygong :: [Q2Dpoint] -> Graphic
polygong = polygon . (map cCg)
--
linef     :: Q2Dpoint -> Q2Dpoint -> Graphic
linef a b = line (cCf a) (cCf b)
--
lineg     :: Q2Dpoint -> Q2Dpoint -> Graphic
lineg a b = line (cCg a) (cCg b)
--
linep     :: Q2Dpoint -> Q2Dpoint -> Graphic
linep a b = line (cCp a) (cCp b)
--
textf     :: Q2Dpoint -> String -> Graphic
textf a t = text (cCf a) t
--
textg     :: Q2Dpoint -> String -> Graphic
textg a t = text (cCg a) t
--
polylinef :: [Q2Dpoint] -> Graphic
polylinef xs = polyline (map cCf xs)
--
polylineg :: [Q2Dpoint] -> Graphic
polylineg xs = polyline (map cCg xs)
--
--
-------------------------------------------------------------------------------
-- grid = draws a grid (coordinates, etc.)
-------------------------------------------------------------------------------
grid :: Graphic
grid = overGraphics [fGrid, gGrid] --, pGrid]
--
fGrid, gGrid, pGrid :: Graphic
fGrid = overGraphics (
  [ linef (fromInt xMin, 0) (fromInt xMax, 0),            -- x axis
    linef (0, fromInt yfMin) (0,fromInt yfMax)            -- y axis
  ] ++
  [ linef (i,-ctyf) (i, ctyf) | i <- xList                -- x subgrid
  ] ++
  [ textf (0, -ctyf) (show 0), textf (1, -ctyf) (show 1)  -- x labels
  ] ++
  [ linef (-ctx, fromInt i) (ctx, fromInt i) | i <- [yfMin .. yfMax]
  ] ++
  [ textf (-10*ctx, fromInt yfMin) (show yfMin),           -- y labels
    textf (-10*ctx, fromInt yfMax) (show yfMax)
  ]
  )
  where
  xList  = map toRational [fromInt xMin, xNext .. fromInt xMax]
  xNext  = fromInt xMin + (fromInt xMax - fromInt xMin)/4
gGrid = overGraphics (
  [ lineg (fromInt xMin, 0) (fromInt xMax, 0),            -- x axis
    lineg (0, fromInt ygMin) (0,fromInt ygMax)            -- y axis
  ] ++
  [ lineg (i,-ctyg) (i, ctyg) | i <- xList                -- x subgrid
  ] ++
  [ textg (0, -ctyg) (show 0), textg (1, -ctyg) (show 1)  -- x labels
  ] ++
  [ lineg (-ctx, fromInt i) (ctx, fromInt i) | i <- [ygMin .. ygMax]
  ] ++
  [ textg (-10*ctx, fromInt ygMin) (show ygMin),          -- y labels
    textg (-10*ctx, fromInt ygMax) (show ygMax)
  ]
  )
  where
  xList  = map toRational [fromInt xMin, xNext .. fromInt xMax]
  xNext  = fromInt xMin + (fromInt xMax - fromInt xMin)/4
pGrid = overGraphics (
  [ linep (fromInt xMin,0) (fromInt xMax,0) ]
  )
--
--
ct, ctx, ctyf,ctyg :: Rational
ct                 = 2%1
ctx                = ct / fromInt  xUnit -- toRational ( xMax -  xMin)
ctyf               = ct / fromInt yfUnit -- toRational (yfMax - ygMin)
ctyg               = ct / fromInt ygUnit -- toRational (ygMax - ygMin)
--
--
-------------------------------------------------------------------------------
-- drawKey - draw key in top left corner
-------------------------------------------------------------------------------
drawKey :: Graphic
drawKey = overGraphics (zipWith drawKey_1 [1..] xs)
  where
  xs = [ (i, keyMapping i) | i <- [11,12,21,30,31,32,33,34,40,41,42,43,44] ]

drawKey_1 :: Int -> (Int, String) -> Graphic
drawKey_1 row (c,v) = overGraphics [drawC, drawV]
  where
  drawC = withRGB (rgb c) (line (10,ypos) (50,ypos))
  drawV = text (60,ypos-8) v
  ypos  = row*16
--
--
-------------------------------------------------------------------------------
-- drawExact - draw an exact function (solution to a problem) with (mesh)
--             points drawn
-------------------------------------------------------------------------------
drawExactf    :: [(Rational,Rational)] -> Graphic
drawExactf fs = withRGB (rgb 99) (polylinef fs)

drawExactg    :: [(Rational,Rational)] -> Graphic
drawExactg fs = withRGB (rgb 199) (polylineg fs)
--
--
-------------------------------------------------------------------------------
--
-------------------------------------------------------------------------------
colourFunDer :: Int
colourFunDer = 100
--
-- INITIAL BOX FUNCTIONS
rgb  11      = RGB 100 100 100  -- Initial Box Function (grayish)
rgb  12      = RGB 200 200 200  -- Recomputed Box Function (grayish)
--
-- VECTOR FIELD
rgb  19      = RGB 150  50  50  -- Cube
rgb  21      = RGB 100   0   0  -- Vector field
--
-- CONSISTENCY
rgb  30      = RGB   0   0 250  -- Consistency (s,t) (light blue)
rgb  31      = RGB 250 250   0  -- Lower Bound for Consistency (s) (yellow)
rgb  32      = RGB 250 150   0  -- Upper Bound for Consistency (t) (orange)
rgb  33      = RGB   0 150   0  -- left 2 right (dark green)
rgb  34      = RGB   0 250   0  -- right 2 left (light green)
--
-- STRONG CONSISTENCY
rgb  40      = RGB   0   0 150  -- Strong Consistency (p,r) (dark blue)
rgb  41      = RGB 250 250 150  -- Lower Bound for Strong Consistency (p) (yellow + noise)
rgb  42      = RGB 250 150 150  -- Upper Bound for Strong Consistency (r) (orange + noise)
rgb  43      = RGB   0 150 150  -- left 2 right (dark green + noise)
rgb  44      = RGB   0 250 150  -- right 2 left (light green + noise)
--
rgb  51      = RGB 250   0   0  --
--
--
rgb 111      = RGB 150 150 150  -- Initial Box Derivative (grayish)
rgb 112      = RGB 200 200 200  -- Recomputed Box Derivative (grayish)
--
--
rgb 119      = RGB 150  50  50  -- Cube
--
--
rgb 151      = RGB 250   0   0  -- 
--
--
rgb   i
  | i <=60               = error "Error in rgb"
  | mod i 3 == 0         = RGB 255 255   0
  | mod i 3 == 1         = RGB 255   0 255
  | mod i 3 == 2         = RGB   0 255 255 -- odd i                = RGB 255 255   0
  | even i               = RGB 255   0 255 --aux2 aux1 aux1
  where
  aux1 = fromInt ((i- 60)*25)
  aux2 = fromInt (255 - (i- 60)*25)
  aux3 = fromInt ((i-160)*25)
  aux4 = fromInt (255 - (i-160)*25)
rgb   i
  | i <=60               = error "Error in rgb"
  | i <= colourFunDer    = RGB aux1 aux1 aux2
  | i  > colourFunDer    = RGB aux3 aux3 aux4
  where
  aux1 = fromInt ((i- 60)*25)
  aux2 = fromInt (255 - (i- 60)*25)
  aux3 = fromInt ((i-160)*25)
  aux4 = fromInt (255 - (i-160)*25)
--
--
rgbNotDef    = RGB 255   0 255  -- Not Defined Regions/Boxes
rgbFail      = RGB 255   0   0  -- 'Something wrong' areas
--
--
--------------------------------------------------------------------------------
-- keyMapping
--------------------------------------------------------------------------------
keyMapping :: Int -> String
keyMapping  11 = "Inital Function"
keyMapping  12 = "Initial Interval Function"
keyMapping  21 = "Vector Field"
keyMapping  30 = "Consistency"
keyMapping  31 = "Consistency (lower)"
keyMapping  32 = "Consistency (upper)"
keyMapping  33 = "Consistency (L2R)"
keyMapping  34 = "Consistency (R2L)"
keyMapping  40 = "Strong Consistency"
keyMapping  41 = "Strong Consistency (lower)"
keyMapping  42 = "Strong Consistency (upper)"
keyMapping  43 = "Strong Consistency (L2R)"
keyMapping  44 = "Strong Consistency (R2L)"
